#include <stdio.h>
#include <sys/stat.h>
#include "rainbowtables.h"


#define BOUND 10
#define TIME_BUF_SIZE 14     // size of buffer for my_timer.timeToString

typedef struct{
        int buffer_len;      // if processing two blocks, each uses half of array
        int full_buffer_len; // if processing only one block, entire array may by used
        int ncol;            //number of collisions
        void *buffer1;
        void *buffer2;
} MergeBlocksInfo;

typedef struct{
        int buffer_len;      // if processing two blocks, each uses half of array
        int full_buffer_len; // if processing only one block, entire array may by used
        void *buffer1;
        void *buffer2;
        int size1, size2;
        
        int ncol;
} MergeFilesInfo;


////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////


// if parameters are ok, returns zero, otherwise returns 1
static int checkParameters(const TableInfo *ti);

// 
static int fileExists(char *file_name);

// adds chains in pFile1 on current position and writes sizes of blocks in pFile1a on current position
static void addChains(FILE *pFile1, FILE *pFile1a, const TableInfo *ti, int chainsnum, void *pArray,
       int *blocksnum);
       
// adds chains in pFile1 on current position and writes sizes of blocks in pFile1a on current position
// but makes some lookups in pFile, to avoid collisions 
static void addChainslookup(FILE *pFile, FILE *pFile1, FILE *pFile1a, const TableInfo *ti, 
       int chainsnum, void *pArray, int *blocksnum, int nlookups, int table_size);

//
static void generateRandomKey(void* key);

// creates chain beginning with random key
static void generateChain(const TableInfo *ti, unsigned char *key, unsigned char *hash);

// compares if entire chain is the same, i.e. if key is the same
static int compareFull(const void* keyhash1, const void* keyhash2);
// compares if chains have merged, i.e. if hash is the same
static int compareHash(const void* keyhash1, const void* keyhash2);

// size1 - size of existing table; size2 - size to be added to the table
static void upgrade(TableInfo *ti, void *pArray, int size1, int size2, int *ncol);

// merges blocks of chains in entire file [ti->table_name]1.tmp; 
// after this function is called, [ti->table_name]1.tmp contains chains sorted according to 
// the compare function in non decreasing order i.e. containing no collisions 
// [ti->table_name]1a.tmp contains only one integer value - number of chains sotred in [ti->table_name]1.tmp
//static int mergeBlocks2(const TableInfo *ti, void *pArray, int blocksnum, int *ncollisions);

static int mergeBlocks(const TableInfo *ti, void *pArray, int blocksnum, int *ncollisions);
       
// merges two consecutive blocks in [file1] at current position and saves them into [file2]
// changes the position to end of second block
static int mergeTwoBlocks(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, MergeBlocksInfo *mbi);

// same as previous, but with removing duplicates
static int mergeTwoBlocksRD(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, MergeBlocksInfo *mbi);

// merges two files 
// reads files from current position, size of each should be stored in MergeFilesInfo
// in mgi->ncol returns the number of collisions
static int mergeTwoFiles(FILE *pFile1, FILE *pFile2, FILE *pFile3, MergeFilesInfo *mfi);

// reads count items from file1, removes collisions, writes to file2
// count is integer stored at current position of pFile1a
static void processBlock(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, void* buffer,
       MergeBlocksInfo *mbi);
       
// same as previous, but with removing duplicates
static void processBlockRD(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, void* buffer,
       MergeBlocksInfo *mbi);

// buffers as much as he can and updates count and pos values; returns number of items buffered
static int bufferBlock(FILE *pFile, void* buffer, int keyhash_sz, int len, int *count, fpos_t *pos);

//
static void initConsts(const TableInfo *ti);

// stores key_sz, hash_sz and chain_length into "<file_name>2.rbt"
static int creatDescriptionFile(const TableInfo *ti);

static int getTableSize(char *fName, int keyhash_sz);

//
static int min(int a, int b);

//
static void resetOutputValues(TableInfo *ti);


// loads ith item from file into *outItem
static void getItem(FILE *pFile, int i, void *outItem);

// chcecks if pFile contains item 
// if yes, returns 1 and position pos is number of that inItem (0..size-1); 
// the value stored in file is copied into outItem
// if no, returns 0 and pos is undefined
static int lookup(FILE *pFile, int table_size, const void *inItem, void *outItem,
       int *pos);

// checks if inItem is stored in table; does not return anything except for true/false
// used for detecting collisions only
// argument outItem is used as local variable, but passsing it from outside makes it faster
static int hashMemberlookup(FILE *pFile, int table_size, const void *inItem, void *outItem);

// chcecks, if key is stored in the table
// needs global_scd and colCompare being set properly
static int keyMemberlookup(FILE *pFile, int table_size, const void *inItem, void *outItem);


void my_fwrite(void* data, FILE *pFile3, void *buffer);

void my_store(FILE *pFile3, void *buffer);

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
// LOCAL VARIABLES
int key_sz, hash_sz, keyhash_sz;
int (*sortCompare)(const void* keyhash1,const void* keyhash2); // table will be sorted acoording to this func.
int (*colCompare)(const void* keyhash1,const void* keyhash2);  // returns zero if items in table .. collision
void (*encode)(void* key, void* hash);
void (*reduce)(void* hash, int i, void* key);
void (*generateKey) (void* key);
 
char str1_tmp[TABLE_NAME_LENGTH + 6]; // table_name + [1 or 2] + ".suf" <- therefore + 6
char str1a_tmp[TABLE_NAME_LENGTH + 6];
char str2_tmp[TABLE_NAME_LENGTH + 6]; // table_name + [1 or 2] + ".suf" <- therefore + 6
char str2a_tmp[TABLE_NAME_LENGTH + 6];
char str[TABLE_NAME_LENGTH + 6];
char str2[TABLE_NAME_LENGTH + 6];
char *file_name1_tmp = str1_tmp;
char *file_name1a_tmp = str1a_tmp;
char *file_name2_tmp = str2_tmp;
char *file_name2a_tmp = str2a_tmp;
char *file_name = str;
char *file_name2 = str2;
unsigned char rnd_buffer[20];    // used for creating chains as result for random function
int global_ncompare;                     // used in lookup function as count of compare calls
int global_ncollisions;
int global_sdd;                         //strict collision detection
//int offset;                      // for formatig output text

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

static int checkParameters(const TableInfo *ti)
{
       if (ti->table_name == "") return 1;
       if (ti->table_size < 1) return 1;
       if (ti->chain_length < 1) return 1;
       if (ti->key_size < 1) return 1;
       if (ti->hash_size < 1) return 1;
       if ((ti->key_size % 8) != 0) return 1;
       if ((ti->hash_size % 8) != 0) return 1;       
       if (ti->fEncode == NULL) return 1;
       if (ti->fReduce == NULL) return 1;
       return 0;
}

static int fileExists(char *file_name)
{
       FILE *pFile;
       int result;
       result = ((pFile = fopen(file_name, "r")) != NULL);
       if (result) fclose(pFile);
       return result;
}


int createTable(TableInfo *ti)
{
    if (checkParameters(ti)) return 4;
    if (ti->array_size < BOUND) { return 1; }
    char timeBuf[TIME_BUF_SIZE];
    
    randomize20(time(NULL));
  	initConsts(ti); 
  	ti->lookups_on = 0;  
    resetOutputValues(ti); 	
    my_time timer_tmp;  	
    void *pArray;   
  	if ( (pArray = malloc(ti->array_size*keyhash_sz)) == NULL ){
           printf("Memory allocation error!");
           return 2;
  	}
  	// Initializing files
  	FILE *pFile1, *pFile1a;
  	if ((pFile1 = fopen(file_name1_tmp, "wb")) == NULL) {
  	   printf("Error while creating table.\nFile cannot be opened %s", file_name1_tmp); 
       return 3;	
    }
  	if ((pFile1a = fopen(file_name1a_tmp, "wb")) == NULL){
  	   printf("Error while creating table.\nFile cannot be opened %s", file_name1a_tmp); 
       return 3;	
    }
    int blocksnum;    
    // Creating table
    printf("Creating Table: %s\n", ti->table_name);
    now(&(ti->create_time));
    addChains(pFile1, pFile1a, ti, ti->table_size, pArray, &blocksnum);    
    now(&(ti->create_chains_time));
    timeDifference(ti->create_time, ti->create_chains_time, &(ti->create_chains_time));
    fclose( pFile1 );
    fclose( pFile1a );
    timeToString(&(ti->create_chains_time), timeBuf);
    printf("  File successfully created in %s seconds.\n", timeBuf);
    printf("  %d chains stored in %d blocks\n\n", ti->table_size, blocksnum);
    
    // now file [table_name]1.tmp contains sorted blocks of chains
    // Merging:
    int ncol; // number of collisions (chains, that have to be added later)
    printf("  Merging blocks...\n");
    mergeBlocks(ti, pArray, blocksnum, &ncol);
    now(&timer_tmp);
    timeDifference(ti->create_time, timer_tmp, &(ti->create_time));
    printf("  Merging done, number of collisions: %d \n", ncol);
    timeToString(&(ti->create_time), timeBuf);
    printf("Time of creating table: %s seconds\n", timeBuf);
    remove(file_name);//if [table_name]SUF exists it will be removed
    rename(file_name1_tmp, file_name);//renames [table_name]1.tmp to [table_name]SUF
    printf("Creating description file... ");
    if (createDescriptionFile(ti)) // stores values into [table_name]2.SUF
       printf("FAILED\n");
    else
       printf("OK\n");    

    int counter = ti->limit1;  //
    ti->upgrade_chains_time.millisec = 0;
    ti->upgrade_chains_time.sec = 0;      
    // if there are any collisions and replacing collisions is on
    if ((ncol)&&(counter)) {
      // if there are any collisions, reset files .tmp 
      // [file_name]1.tmp will be later merged with [file_name]SUF
       if ((pFile1 = fopen(file_name1_tmp, "wb")) == NULL){  //resets file_name1
  	      printf("Error while creating table.\nFile cannot be opened %s", file_name1_tmp); 
          return 3;	
       } 
       fclose(pFile1);
       if ((pFile1a = fopen(file_name1a_tmp, "wb")) == NULL){
  	      printf("Error while creating table.\nFile cannot be opened %s", file_name1a_tmp); 
          return 3;	
       }
       fclose(pFile1a);
    } else {
      ti->upgrade_time.millisec = 0;
      ti->upgrade_time.sec = 0;      
      remove(file_name1a_tmp);
      free (pArray);
      ti->collisions = ncol;
      ti->chains = ti->table_size - ncol;
      return 0;
    }
    
    int size1 = ti->table_size - ncol; // size of table that was already created
    int size2 = ncol;                  // size to be added to the table
//    printf("Upgrade size1:%d size2:%d\n", size1, size2);
    int ncol2 = ncol;
    printf("Replacing collisions...\n");
    now(&(ti->upgrade_time));
    while ((ncol)&&(counter--)){
          upgrade(ti, pArray, size1, size2, &ncol);
          size1 += size2 - ncol;
          size2 = ncol;
          ncol2 += global_ncollisions;
    }
    now(&timer_tmp);
    timeDifference(ti->upgrade_time, timer_tmp, &(ti->upgrade_time));
    timeToString(&(ti->upgrade_time), timeBuf);
    printf("Replacing done!\n");
    printf("Time of replacig collisions: %s seconds\n", timeBuf);
    ti->collisions = ncol2;
    ti->chains = ti->table_size - ncol; 
    free(pArray);
  	return 0;
}

static void addChains(FILE *pFile1, FILE *pFile1a, const TableInfo *ti, int chainsnum, void *pArray,
       int *blocksnum)
{       
       printf("  Creating %d chains\n", chainsnum);
     	 int i, offset;
       int count; // length of current block
       *blocksnum = 0; //number of blocks to be merged together
       while (chainsnum>0){           
             count = min(ti->array_size, chainsnum); // count = number of chains in block
             fwrite(&count, sizeof(int), 1, pFile1a);
             offset = 0;
             for(i=0; i<count; i++){
                      generateChain(ti, (pArray + offset), (pArray + offset + key_sz));
                      offset += keyhash_sz;
             }
             (*blocksnum)++;
             chainsnum -= count;
             qsort(pArray, count, keyhash_sz, sortCompare);
             fwrite(pArray, keyhash_sz, count, pFile1);
       }
}

static void addChainslookup(FILE *pFile, FILE *pFile1, FILE *pFile1a, const TableInfo *ti, 
       int chainsnum, void *pArray, int *blocksnum, int nlookups, int table_size)
{
       printf("  Creating %d chains\n", chainsnum);
       printf("LOOKUP %d ts: %d\n", nlookups, table_size);
       unsigned char keyhash[keyhash_sz]; // argument for lookup (inItem)
     	 int i, offset;
       int count; // length of current block
       *blocksnum = 0; //number of blocks to be merged together
       int l = 0;
       while (chainsnum>0){           
             count = min(ti->array_size, chainsnum); // count = number of chains in block
             fwrite(&count, sizeof(int), 1, pFile1a);
             offset = 0;
             for(i=0; i<count; i++){
                      generateChain(ti, (pArray + offset), (pArray + offset + key_sz));
                      if ((nlookups)&& (keyMemberlookup(pFile, table_size, pArray + offset, keyhash))) {
                         i--;
                         nlookups--;
                         l++;
                      } else {
                         offset += keyhash_sz;
                      }
             }
             (*blocksnum)++;
             chainsnum -= count;
             qsort(pArray, count, keyhash_sz, sortCompare);
             fwrite(pArray, keyhash_sz, count, pFile1);
       }
       printf("Chains Thrown Away: %d\n", l);
}

static void generateRandomKey(void* key)
{
       randomN(key, key_sz);
}

static void generateChain(const TableInfo *ti, unsigned char *key, unsigned char *hash)
{
       int i;
       unsigned char k[key_sz];
       
       generateKey(key);
       
       memcpy(k, key, key_sz);
       
       for (i=1; i<ti->chain_length; i++){ // ta 1 je tam naschval
           encode(k, hash);
           reduce(hash, i, k);
       }
       encode(k, hash);
       return;       
}

static int compareFull(const void* keyhash1, const void* keyhash2)
{
       return memcmp(keyhash1, keyhash2, key_sz);
}

static int compareHash(const void* keyhash1, const void* keyhash2)
{
       return memcmp(keyhash1 + key_sz, keyhash2 + key_sz, hash_sz);
}

static void upgrade(TableInfo *ti, void *pArray, int size1, int size2, int *ncol)
{
//      printf("UPGRADE col:%d\n", *ncol);
//      printf("UPGRADE size1:%d size2:%d\n", size1, size2);
      printf("Upgrading table....\n");
      printf("  Table_size: %d\n", size1);
      printf("  Chains to be added: %d\n", size2);
      
      my_time timer1, timer2;
      
    	FILE *pFile1, *pFile1a, *pFile;
      MergeFilesInfo mfi;
      int blocksnum;
      mfi.size1 = size1;
      mfi.buffer_len = ti->array_size / 2;
      mfi.full_buffer_len = ti->array_size;
      mfi.buffer1 = pArray;
      mfi.buffer2 = pArray + (keyhash_sz*mfi.buffer_len); // pointer on second block
      
      global_ncollisions = 0;
      *ncol = size2;
      int _ncol = size2; 
      int counter = ti->limit2;
      int other_blocks=0; // non zero value means there is one more block in file pFile1 to be merged
      while ((*ncol)&&(counter--)) {
          	if ((pFile1 = fopen(file_name1_tmp, "ab")) == NULL) {
  	           printf("Error while upgrading.\nFile cannot be opened %s", file_name1_tmp); 
  	           return;
            }
            if ((pFile1a = fopen(file_name1a_tmp, "ab")) == NULL){
  	           printf("Error while upgrading.\nFile cannot be opened %s", file_name1a_tmp); 
  	           return;
            }
            	  
            if (ti->lookups_on){
            	 if ((pFile = fopen(file_name, "rb")) == NULL){
                  printf("Error while upgrading.\nFile cannot be opened %s", file_name); 
  	              return;
               }
               now(&timer1);
               addChainslookup(pFile, pFile1, pFile1a, ti, *ncol, pArray, &blocksnum, ti->lookups, mfi.size1);
               now(&timer2);
               timeDifference(timer1, timer2, &timer2);
               timeSum(&(ti->upgrade_chains_time), timer2);
               fclose( pFile1 );
               fclose( pFile1a );
               fclose( pFile );
               printf("  %d chains created in %d blocks\n", *ncol, blocksnum);
               if (other_blocks) {blocksnum++;} else {other_blocks=1;}
               printf("  Merging blocks....");
               mergeBlocks(ti, pArray, blocksnum, ncol);
               printf("DONE.\n");
            } else {
               now(&timer1);
               addChains(pFile1, pFile1a, ti, *ncol, pArray, &blocksnum);
               now(&timer2);
               timeDifference(timer1, timer2, &timer2);
               timeSum(&(ti->upgrade_chains_time), timer2);
               fclose( pFile1 );
               fclose( pFile1a );
               printf("  %d chains created in %d blocks\n", *ncol, blocksnum);
               if (other_blocks) {blocksnum++;} else {other_blocks=1;}
               printf("  Merging blocks....");
               mergeBlocks(ti, pArray, blocksnum, ncol);
               printf("DONE.\n");
            }
            global_ncollisions += *ncol;
            if ((*ncol == 0)||(counter == 0)) {               
               other_blocks=0;
               if ((pFile1 = fopen(file_name1_tmp, "rb")) == NULL){
  	              printf("Error while upgrading.\nFile cannot be opened %s", file_name1_tmp); 
  	              return;
               }
               if ((pFile = fopen(file_name, "rb")) == NULL){
  	              printf("Error while upgrading.\nFile cannot be opened %s", file_name);
  	              return;
               }
               // this file will be used as output for merged files [table_name]SUF and [table_name]1.tmp
               if ((pFile1a = fopen(file_name1a_tmp, "wb")) == NULL){
  	              printf("Error while upgrading.\nFile cannot be opened %s", file_name1a_tmp); 
  	              return;
               }
         	     mfi.size2 = _ncol - *ncol; //_ncol chains were added, *ncol were dropped in mergeBlocks
         	     printf("  %d chains successfully created\n", _ncol);
//         	   printf("size1: %d, size2: %d\n", mfi.size1, mfi.size2);
               printf("  Merging original table and new table...\n"); 
               mergeTwoFiles(pFile, pFile1, pFile1a, &mfi);
               printf("  Merging done, number of collisions: %d\n", mfi.ncol);  
               global_ncollisions += mfi.ncol;
               
               *ncol = _ncol = mfi.ncol;
               mfi.size1 += mfi.size2 - mfi.ncol;  
               fclose(pFile1);
               fclose(pFile1a);
               fclose(pFile);
               remove(file_name);//if [table_name]SUF exists it will be removed
               rename(file_name1a_tmp, file_name);//renames [table_name]1a.tmp to [table_name]SUF
               remove(file_name1_tmp);               
            }
      }
      printf("Upgrading Done!\n");   
}

static int mergeBlocks(const TableInfo *ti, void *pArray, int blocksnum, int *ncollisions)
{
       MergeBlocksInfo mbi;
       mbi.buffer_len = ti->array_size / 2;
       mbi.full_buffer_len = ti->array_size;
       mbi.buffer1 = pArray; //pointer on first block
       mbi.buffer2 = pArray + (keyhash_sz*mbi.buffer_len); // pointer on second block
       mbi.ncol = 0;
       unsigned char key[key_sz];   
       unsigned char hash[hash_sz];
       

       FILE *pInFile, *pOutFile, *pInaFile, *pOutaFile;
       
       if (blocksnum==1){
           if ((pInFile = fopen(file_name1_tmp, "rb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name1_tmp); 
              return;
           }
           if ((pInaFile = fopen(file_name1a_tmp, "rb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name1a_tmp); 
              return;
           }
           if ((pOutFile = fopen(file_name2_tmp, "wb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name2_tmp); 
              return;
           }
           if ((pOutaFile = fopen(file_name2a_tmp, "wb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name2a_tmp); 
              return;
           }
           processBlockRD(pInFile, pInaFile, pOutFile, pOutaFile, pArray, &mbi);
           *ncollisions = mbi.ncol;
           fclose(pInFile);
           fclose(pInaFile);
           fclose(pOutFile);
           fclose(pOutaFile);
           remove(file_name1_tmp);
           remove(file_name1a_tmp);
           rename(file_name2_tmp, file_name1_tmp);
           rename(file_name2a_tmp, file_name1a_tmp);
           return 0;
       }
       int n;
       // while there are block to be merged or there is only one block with duplicates
       while (blocksnum>1){
//           printf("MERGING %d blocks\n", blocksnum);
           if ((pInFile = fopen(file_name1_tmp, "rb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name1_tmp); 
              return;
           }
           if ((pInaFile = fopen(file_name1a_tmp, "rb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name1a_tmp); 
              return;
           }
           if ((pOutFile = fopen(file_name2_tmp, "wb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name2_tmp); 
              return;
           }
           if ((pOutaFile = fopen(file_name2a_tmp, "wb")) == NULL){
              printf("Error while merging blocks.\nFile cannot be opened %s", file_name2a_tmp); 
              return;
           }
           n = (blocksnum / 2) - (blocksnum % 2);
           if (ti->remove_duplicates) {
              while ((n--) >0)
                    mergeTwoBlocksRD(pInFile, pInaFile, pOutFile, pOutaFile, &mbi);
              if (blocksnum % 2) {
                 processBlockRD(pInFile, pInaFile, pOutFile, pOutaFile, pArray, &mbi);           
                 mergeTwoBlocksRD(pInFile, pInaFile, pOutFile, pOutaFile, &mbi);
              }
           } else {
             if (blocksnum==2) {
                mergeTwoBlocksRD(pInFile, pInaFile, pOutFile, pOutaFile, &mbi);
                blocksnum = 1;
             } else {
                while ((n--) >0)
                      mergeTwoBlocks(pInFile, pInaFile, pOutFile, pOutaFile, &mbi);
                if (blocksnum % 2) {
                   processBlock(pInFile, pInaFile, pOutFile, pOutaFile, pArray, &mbi);           
                   mergeTwoBlocks(pInFile, pInaFile, pOutFile, pOutaFile, &mbi);
                }
             }
           }
           fclose(pInFile);
           fclose(pInaFile);
           fclose(pOutFile);
           fclose(pOutaFile);
           blocksnum = (blocksnum+1)/2;
           remove(file_name1_tmp);
           remove(file_name1a_tmp);
           rename(file_name2_tmp, file_name1_tmp);
           rename(file_name2a_tmp, file_name1a_tmp);
       }
       *ncollisions = mbi.ncol;
       return 0;
}

static int mergeTwoBlocks(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, MergeBlocksInfo *mbi)
{
       int count1, count2; // number of items to be merged from ith block (i = 1 or 2)
       int nbuf1 = 0;// number of items in 1st buffer
       int nbuf2 = 0; 
       unsigned char *buffer1 = mbi->buffer1;
       unsigned char *buffer2 = mbi->buffer2;      
       
       int c; // result of compare;
       fpos_t pos1, pos2; // position in [file1]; points on first item in ith block than was not buffered
       fpos_t end_of_second_block; // this will be set as position in pFile1 at the end
       fread(&count1, sizeof( int ), 1, pFile1a );
       pos1 = ftell(pFile1);
       pos2 = pos1 + (count1*keyhash_sz); 
       fread(&count2, sizeof( int ), 1, pFile1a );
       int count = count1 + count2;
       int ncol = 0;
       end_of_second_block = pos2 + (count2*keyhash_sz);
       unsigned char *i1 = buffer1; //points on current item in first block
       unsigned char *i2 = buffer2;
       
       while ((nbuf1 || (nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mbi->buffer_len, &count1, &pos1)))&&
              (nbuf2 || (nbuf2 = bufferBlock(pFile1, buffer2, keyhash_sz, mbi->buffer_len, &count2, &pos2)))  ) {
             c = sortCompare(i1, i2);
             if ((c==0)&&(!global_sdd)) 
                c=compareFull(i1, i2);
             if (c>0) {
                // first is bigger, so second will be stored
                fwrite(i2, keyhash_sz, 1, pFile2);
                nbuf2--;
                i2 += keyhash_sz;
             } else
             if (c<0) {
                // second is bigger, so first will be stored
                fwrite(i1, keyhash_sz, 1, pFile2);
                nbuf1--;
                i1 += keyhash_sz;                               
             } else {
                //if both items are equal
                nbuf2--;
                i2 += keyhash_sz;
                ncol++;
             }
             if (nbuf1==0) i1 = buffer1;
             if (nbuf2==0) i2 = buffer2;
       }
       //now one of the buffers is empty (the other is not);
       if (nbuf1==0){
          // if first block is processed but second is not
          i1 = i2;
          nbuf1 = nbuf2;
          count1 = count2;
          pos1 = pos2;
       }
       fwrite(i1, keyhash_sz, nbuf1, pFile2);
       nbuf1 = 0;
       i1 = buffer1;
       while (nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mbi->full_buffer_len, &count1, &pos1))
             fwrite(i1, keyhash_sz, nbuf1, pFile2);
       fsetpos(pFile1, &end_of_second_block);
       
       count -= ncol;
       mbi->ncol += ncol;
       fwrite(&count, sizeof(int), 1, pFile2a);
}

static int mergeTwoBlocksRD(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, MergeBlocksInfo *mbi)
{
       int count1, count2; // number of items to be merged from ith block (i = 1 or 2)
       int nbuf1 = 0;// number of items in 1st buffer
       int nbuf2 = 0; 
       unsigned char *buffer1 = mbi->buffer1;
       unsigned char *buffer2 = mbi->buffer2;      
       
       unsigned char llast[keyhash_sz]; // last stored value; for detecting collisions;
       unsigned char *last = llast;
       int c; // result of compare;
       fpos_t pos1, pos2; // position in [file1]; points on first item in ith block than was not buffered
       fpos_t end_of_second_block; // this will be set as position in pFile1 at the end
       fread(&count1, sizeof( int ), 1, pFile1a );
       pos1 = ftell(pFile1);
       pos2 = pos1 + (count1*keyhash_sz); 
       fread(&count2, sizeof( int ), 1, pFile1a );
       int count = count1 + count2; // merged block will contain count-ncol items
       end_of_second_block = pos2 + (count2*keyhash_sz);
       unsigned char *i1 = buffer1; //points on current item in first block
       unsigned char *i2 = buffer2;
       int ncol = 0;
       
       nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mbi->buffer_len, &count1, &pos1);
       nbuf2 = bufferBlock(pFile1, buffer2, keyhash_sz, mbi->buffer_len, &count2, &pos2);
       // first colCompare is always false
       last[0] = i1[0] + 1;
       if (last[0]==i2[0]) last[0]++;
       last[keyhash_sz-1] = i2[keyhash_sz-1] + 1;
       if (last[keyhash_sz-1]==i2[keyhash_sz-1]) last[keyhash_sz-1]++;
       
       do {
          c = sortCompare(i1, i2);
          if ((c==0)&&(!global_sdd)) 
             c=colCompare(i1, i2);
          if (c>0) {
             // first is bigger, so second will be stored
             if (colCompare(i2, &last)){
                // i2 points on diferent chain than last
                fwrite(i2, keyhash_sz, 1, pFile2);
                last = i2;
             } else {
                ncol++;
             }
             nbuf2--;
             if (!nbuf2){ 
                memcpy(llast, last, keyhash_sz);
                last = llast;
             }
             i2 += keyhash_sz;
          } else
          if (c<0) {
             // second is bigger, so first will be stored
             if (colCompare(i1, &last)){
                // i1 points on diferent chain than last
                fwrite(i1, keyhash_sz, 1, pFile2);
                last = i1;
             } else {
                ncol++;
             }
             nbuf1--;
             if (!nbuf1){ 
                memcpy(llast, last, keyhash_sz);
                last = llast;
             }
             i1 += keyhash_sz;                               
          } else {
             //if both items are equal
             ncol++;
             nbuf2--;
             i2 += keyhash_sz;
             if (!nbuf2){ 
                memcpy(llast, last, keyhash_sz);
                last = llast;
             }
          }
          if (nbuf1==0) i1 = buffer1;
          if (nbuf2==0) i2 = buffer2;
       } while ( (nbuf1 || (nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mbi->buffer_len, &count1, &pos1)))&&
                 (nbuf2 || (nbuf2 = bufferBlock(pFile1, buffer2, keyhash_sz, mbi->buffer_len, &count2, &pos2)))  );
       //now one of the buffers is empty (the other is not);
       if (nbuf1==0){
          // if first block is processed but second is not
          i1 = i2;
          nbuf1 = nbuf2;
          count1 = count2;
          pos1 = pos2;
       }
       fwrite(i1, keyhash_sz, nbuf1, pFile2);
       while (nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mbi->full_buffer_len, &count1, &pos1))
             fwrite(buffer1, keyhash_sz, nbuf1, pFile2);
             
       fsetpos(pFile1, &end_of_second_block);
       mbi->ncol += ncol;
       count -= ncol;
       fwrite(&count, sizeof(int), 1, pFile2a);
       fsetpos(pFile1, &end_of_second_block);
}

static int mergeTwoFiles(FILE *pFile1, FILE *pFile2, FILE *pFile3, MergeFilesInfo *mfi)
{
       int nbuf1 = 0;// number of items in 1st buffer
       int nbuf2 = 0; 
       unsigned char *buffer1 = mfi->buffer1;
       unsigned char *buffer2 = mfi->buffer2;                    
       int c; //result of compare;
       fpos_t pos1, pos2; // position in [file1];
       pos1 = ftell(pFile1);
       pos2 = ftell(pFile2);
       unsigned char *i1 = buffer1;
       unsigned char *i2 = buffer2;
       
       int count1 = mfi->size1;
       int count2 = mfi->size2;
       int ncol = 0;
       
       while ( (nbuf1 || (nbuf1 = bufferBlock(pFile1, buffer1, keyhash_sz, mfi->buffer_len, &count1, &pos1)))&&
               (nbuf2 || (nbuf2 = bufferBlock(pFile2, buffer2, keyhash_sz, mfi->buffer_len, &count2, &pos2)))  ){
             c = sortCompare(i1, i2);
             // if they are equal and scd -> throw newly generated away
             //                if not scd -> decision based on compareFull
             if ((c==0)&&(!global_sdd))
                c = compareFull(i1, i2) ;
             if (c>0) {
                // first is bigger, so second will be stored
                fwrite(i2, keyhash_sz, 1, pFile3);
                nbuf2--;
                i2 += keyhash_sz;
             } else
             if (c<0) {
                // second is bigger, so first will be stored
                // here is no compare, since first file has to be merged whole
                fwrite(i1, keyhash_sz, 1, pFile3);
                nbuf1--;
                i1 += keyhash_sz;                               
             } else {
                //if both items are equal, throw the second away
                ncol++;
                nbuf2--;
                i2 += keyhash_sz;
             }
             if (nbuf1==0) i1 = buffer1;
             if (nbuf2==0) i2 = buffer2;
       }
       
       //now one of the buffers is empty;
       FILE *pFile;
       if (nbuf1==0) {
          // if first block is empty,                      
          i1 = i2;
          nbuf1 = nbuf2;
          count1 = count2;
          pos1 = pos2;
          pFile = pFile2;
       } else {
          pFile = pFile1;         
       }
       fwrite(i1, keyhash_sz, nbuf1, pFile3);
       while (nbuf1 = bufferBlock(pFile, buffer1, keyhash_sz, mfi->full_buffer_len, &count1, &pos1))
             fwrite(buffer1, keyhash_sz, nbuf1, pFile3);
       mfi->ncol = ncol;
}

static void processBlock(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, void* buffer,
       MergeBlocksInfo *mbi)
{
       int count;
       fread(&count, sizeof(int), 1, pFile1a);
       fwrite(&count, sizeof(int), 1, pFile2a);
       fpos_t pos;
       pos = ftell(pFile1);
       int nbuf = 0;
       int ncol = 0;
       while (nbuf = bufferBlock(pFile1, buffer, keyhash_sz, mbi->full_buffer_len, &count, &pos))
             fwrite(buffer, keyhash_sz, nbuf, pFile2);
}

static void processBlockRD(FILE *pFile1, FILE *pFile1a, FILE *pFile2, FILE *pFile2a, void* buffer,
       MergeBlocksInfo *mbi)
{
       unsigned char llast[keyhash_sz]; // last stored value; to detect collisions;
       unsigned char *last = llast;
       int c;
       for (c=0; c<keyhash_sz; c++)
           last[0] = 0;
       int count, count1;
       fread(&count, sizeof(int), 1, pFile1a);
//       printf("pb: size: %d\n", count);
       count1 = count;
       fpos_t pos;
       pos = ftell(pFile1);
       int nbuf = 0;
       int ncol = 0;
       unsigned char *i = buffer;
         //TODO tu by to chcelo vetvenie, kvoli tomu, ci robim vyhadzovanie kolizii teraz, alebo az na konci
       
       nbuf = bufferBlock(pFile1, buffer, keyhash_sz, mbi->full_buffer_len, &count, &pos);
       last[0] = i[0] + 1;// first colCompare is always false
       last[keyhash_sz-1] = i[keyhash_sz-1] + 1;// first colCompare is always false
       do {
          if (colCompare(i, &last)){
             fwrite(i, keyhash_sz, 1, pFile2);
             if (nbuf == 1) {
                last = llast;
                memcpy(last, i, keyhash_sz);
             } else
                last = i;
          } else {
             ncol++;
             if (nbuf==1) {
                memcpy(llast, last, keyhash_sz);
                last = llast;
             }
          }
          nbuf--;
          i += keyhash_sz;                               
          if (nbuf==0) i = buffer;
       }  while (nbuf || (nbuf = bufferBlock(pFile1, buffer, keyhash_sz, mbi->full_buffer_len, &count, &pos)));
       
       mbi->ncol += ncol;
       count1 -= ncol;
       fwrite(&count1, sizeof(int), 1, pFile2a);
}

static int bufferBlock(FILE *pFile, void* buffer, int keyhash_sz, int len, int *count, fpos_t *pos)
{
       int res; //result: number of intems buffered;
       if (! (res = min(len, *count))) return 0;  // min of (buffer length, chians left)
       *count -= res;                             // chains left -= now buffered
       fsetpos(pFile, pos);
       fread(buffer, keyhash_sz, res, pFile);
       *pos = ftell(pFile);
       return res;
}

static void initConsts(const TableInfo *ti)
{              
       strcpy(file_name1_tmp, ti->table_name);
       strcat(file_name1_tmp, "1.tmp");
       
       strcpy(file_name1a_tmp, ti->table_name);
       strcat(file_name1a_tmp, "1a.tmp");
       
       strcpy(file_name2_tmp, ti->table_name);
       strcat(file_name2_tmp, "2.tmp");
       
       strcpy(file_name2a_tmp, ti->table_name);
       strcat(file_name2a_tmp, "2a.tmp");
       
       strcpy(file_name, ti->table_name);
       strcat(file_name, SUF);
       
       strcpy(file_name2, ti->table_name);
       strcat(file_name2, "2");       
       strcat(file_name2, SUF);
       
       key_sz = ti->key_size / 8;  //value in bytes
       hash_sz = ti->hash_size / 8;
       keyhash_sz = key_sz + hash_sz;
       reduce = ti->fReduce;
       encode = ti->fEncode;
       
       if (ti->fCompare == NULL){
          sortCompare = &compareHash;          
       } else {
          sortCompare = ti->fCompare;
       }
       
       global_sdd = ti->strict_dupl_detection;
       if (ti->strict_dupl_detection) {
         colCompare = &compareHash;
       } else {
         colCompare = &compareFull;
       }
       
       if (ti->fGenerateKey == NULL){
          generateKey = &generateRandomKey;
       } else {
          generateKey = ti->fGenerateKey;
       }
}

int createDescriptionFile(const TableInfo *ti)
{
    	 FILE *pFile;
    	 if ((pFile = fopen(file_name2, "wb")) == NULL) {
          printf("Error while crating description file.");
          return 1;
       }
       fwrite(&(ti->key_size), sizeof(int), 1, pFile);
       fwrite(&(ti->hash_size), sizeof(int), 1, pFile);
       fwrite(&(ti->chain_length), sizeof(int), 1, pFile);
       fclose(pFile);
 	     return 0;
}

int loadMainParameters(TableInfo *ti, int loadTableSize)
{
    FILE *pFile;
    char str[TABLE_NAME_LENGTH + 6];
    char *fName = str;
    char str2[TABLE_NAME_LENGTH + 6];
    char *fName2 = str2;
    int result = 0;
    // check if table exists
    strcpy(fName, ti->table_name);
    strcat(fName, SUF);// fName = [table_name].SUF
 	  if (! fileExists(fName)) result++;
       
    strcpy(fName2, ti->table_name);
    strcat(fName2, "2");
    strcat(fName2, SUF);// fName = [table_name]2.SUF

 	  if ((pFile = fopen(fName2, "rb")) == NULL) {
       result += 2;
       return result;
       if ((loadTableSize)&&(result==2)) {
          //this will only by proceeded if table exists;
          ti->table_size = getTableSize(fName, (ti->key_size + ti->hash_size)/8);
       }
    }
    
    fread(&(ti->key_size), sizeof(int), 1, pFile);
    fread(&(ti->hash_size), sizeof(int), 1, pFile);
    fread(&(ti->chain_length), sizeof(int), 1, pFile);
    
    if ((loadTableSize)&&(!result)) {
       //this will only by proceeded if both files exist;
       ti->table_size = getTableSize(fName, (ti->key_size + ti->hash_size)/8);
    }
    
    fclose(pFile);
    
    return result;
}


static int getTableSize(char *fName, int keyhash_sz)
{
       struct stat file_status;
       stat(fName, &file_status);
       return file_status.st_size / keyhash_sz;
}

static int min(int a, int b)
{
       return (a<b) ?  a : b;
}

static void resetOutputValues(TableInfo *ti)
{
       ti->row = ti->column = -1;
       ti->false_alarms = 0;
       ti->nencode = 0;
       ti->false_nencode = 0;
       ti->nreduce = 0;
       ti->false_nreduce = 0;
       ti->ncompare = 0;
       
       resetTime(&(ti->create_time));
       resetTime(&(ti->create_chains_time));
       resetTime(&(ti->upgrade_time));
       resetTime(&(ti->upgrade_chains_time));
       resetTime(&(ti->search_time));
}

static int lookup(FILE *pFile, int table_size, const void *inItem, void *outItem,
       int *pos)
{
       // we are looking for a sequence of items in the table,  which have the same [hash] part
       // variables l, h represent bounds for the beginning of this sequence
       // variables ll,hh are bounds for end of this sequence
       
       // l, h, ll, hh are now set so they contain the whole table
       int l = 0; // 
       int h = table_size - 1;
       int ll = 0;
       int hh = h;
       int c;  // result of compare
       int i;  // (l+h)/2
       fpos_t fpos; // position in file pFile
       global_ncompare = 0;
       int seeking = 1;
       while ((l<=h)&&(seeking)){
             i = (l+h) / 2;
//             printf("l: %d, h: %d, i: %d\n", l, h, i);
             fpos = i*keyhash_sz;
             fsetpos(pFile, &fpos);
             fread(outItem, keyhash_sz, 1, pFile);
             c = sortCompare(inItem, outItem);
             global_ncompare++;
             if (c>0){
                l = i+1;
                ll = l;
             } else
             if (c<0){
                h = i-1;
                hh = h;
             } else {
               seeking = 0;
               h = i;
               ll = i;
             }
       }
       while (l<h){
             i = (l+h) / 2;
             fpos = i*keyhash_sz;
             fsetpos(pFile, &fpos);
             fread(outItem, keyhash_sz, 1, pFile);
             c = sortCompare(inItem, outItem);
             global_ncompare++;
             if (c>0){
                l = i+1;
             } else
             if (c<0){
                h = i-1;
             } else {
               h = i;
             }
       }
       if (seeking) return 0;
       while (ll<hh){
             i = (ll+hh+1) / 2;
//             printf("%d\n", i);
             fpos = i*keyhash_sz;
             fsetpos(pFile, &fpos);
             fread(outItem, keyhash_sz, 1, pFile);
             c = sortCompare(inItem, outItem);
             global_ncompare++;
             if (c>0){
                ll = i+1;
             } else
             if (c<0){
                hh = i-1;
             } else {
               ll = i;
             }
       }
       *pos = l;
       // now hh=ll; h=l
       return hh-l+1;
       printf("OK");
}

static int hashMemberlookup(FILE *pFile, int table_size, const void *inItem, void *outItem)
{
       int l = 0;
       int h = table_size - 1;
       int c;  // result of compare
       int i;  // (l+h)/2
       fpos_t fpos;
       int j;
       
       while (l<=h){
             i = (l+h) / 2;
             
             fpos = i*keyhash_sz;
             fsetpos(pFile, &fpos);
             fread(outItem, keyhash_sz, 1, pFile);
             
             c = sortCompare(inItem, outItem);
             if (c>0){
                l = i+1;
             } else
             if (c<0){
                h = i-1;
             } else {
               return 1;
             }
       }
       return 0;
}

static int keyMemberlookup(FILE *pFile, int table_size, const void *inItem, void *outItem)
{
       int pos;
       int items_found;
       if (global_sdd) {
          return hashMemberlookup(pFile, table_size, inItem, outItem);        
       } else {
          //if strict_dupl_detection is off;
          //check for keys          
          items_found = lookup(pFile, table_size, inItem, outItem, &pos);
       }
       int k;
       for (k=0; k<items_found; k++){
           getItem(pFile, pos, outItem);
           if (compareFull(inItem, outItem)==0) return 1;
           pos++;
       }
       return 0;
}


void saveToTextfile(const TableInfo *ti, char *filename)
{
     FILE *pFile;
	   if ((pFile = fopen(filename, "w")) == NULL){
   	    printf("Error! File cannot be opened %s", file_name); 
        return;	
     }
	   
	   unsigned char timeBuf[TIME_BUF_SIZE];
	   
	   fprintf( pFile, "Table name: \"%s\"\n\n", ti->table_name);
	   fprintf( pFile, "Key size             = %d bits\n", ti->key_size);
	   fprintf( pFile, "Hash size            = %d bits\n", ti->hash_size);
	   fprintf( pFile, "Table size           = %d\n", ti->table_size);
	   fprintf( pFile, "Length of chain      = %d\n", ti->chain_length);
	   fprintf( pFile, "Array size           = %d\n", ti->array_size);
	   fprintf( pFile, "Remove duplicates    : ");
	   if (ti->remove_duplicates) 
        fprintf( pFile, "On\n"); 
     else 
        fprintf( pFile, "Off\n");
	   fprintf( pFile, "Strict col. detection: ");
	   if (ti->strict_dupl_detection) 
        fprintf( pFile, "On\n"); 
     else 
        fprintf( pFile, "Off\n");
	   fprintf( pFile, "Lookups              : ");
	   if (ti->lookups_on) 
        fprintf( pFile, "On\n"); 
     else 
        fprintf( pFile, "Off\n");
	   fprintf( pFile, "Number of lookups    = %d\n", ti->lookups);
	   fprintf( pFile, "Limit1               = %d\n", ti->limit1);
	   fprintf( pFile, "Limit2               = %d\n\n", ti->limit2);	   
	   fprintf( pFile, "Outpus values:\n");
	   fprintf( pFile, "Number of collisions = %d\n", ti->collisions);
	   fprintf( pFile, "Chains added         = %d\n", ti->chains);
	   fprintf( pFile, "False alarms         = %d\n", ti->false_alarms);
	   fprintf( pFile, "NEncode              = %d\n", ti->nencode);
	   fprintf( pFile, "FNencode             = %d\n", ti->false_nencode);
	   fprintf( pFile, "NReduce              = %d\n", ti->nreduce);
	   fprintf( pFile, "FNReduce             = %d\n", ti->false_nreduce);
	   fprintf( pFile, "NCompare             = %d\n", ti->ncompare);
     timeToString(&(ti->create_time), timeBuf);
	   fprintf( pFile, "Creating table time             = %s seconds\n", timeBuf);
     timeToString(&(ti->create_chains_time), timeBuf);
	   fprintf( pFile, "Creating chains time            = %s seconds\n", timeBuf);
     timeToString(&(ti->upgrade_time), timeBuf);
	   fprintf( pFile, "Upgrading table time            = %s seconds\n", timeBuf);
     timeToString(&(ti->upgrade_chains_time), timeBuf);
	   fprintf( pFile, "Creating chains while upgrading = %s seconds\n", timeBuf);
     timeToString(&(ti->search_time), timeBuf);
	   fprintf( pFile, "Searching time                  = %s seconds\n", timeBuf);
	   fclose(pFile);
}

void saveToTextfile2(const TableInfo *ti, char *filename, char *open_type, void *item)
{
     if ((! strcmp(open_type, "w"))&&(! strcmp(open_type, "a"))) return;
     FILE *pFile;
     int first = (strcmp(open_type, "w"));
     int i;
	   if ((pFile = fopen(filename, open_type)) == NULL)
   	   { perror( "fopen" ); return;	}
	   
	   if (!first) {
    	  fprintf( pFile, "Project name: %s\n\n", ti->table_name);
    	  fprintf( pFile, "Key size        = %d bits\n", ti->key_size);
    	  fprintf( pFile, "Hash size       = %d bits\n", ti->hash_size);
    	  fprintf( pFile, "Table size      = %d\n", ti->table_size);
  	    fprintf( pFile, "Length of chain = %d\n", ti->chain_length);
     }
	   
	   fprintf( pFile, "\n\nLooking for: ");
	   for (i=0; i<((ti->hash_size)/8); i++)
         fprintf( pFile, "%02x ", ((unsigned char *)(item))[i]);
     if (ti->row >= 0){
        fprintf( pFile, "     Key found\n");
        fprintf( pFile, "Key            = ");
	      for (i=0; i<((ti->key_size)/8); i++)
            fprintf(pFile, "%02x ", ((unsigned char *)(ti->pKey))[i]);
        fprintf( pFile, "\n\n" );
        fprintf( pFile, "Row            = %d\n", ti->row );
        fprintf( pFile, "Column         = %d\n", ti->column );
     } else
	      fprintf( pFile, "     Key not found\n");    
	   fprintf( pFile, "False alarms   = %d\n", ti->false_alarms);
	   fprintf( pFile, "NEncode        = %d\n", ti->nencode);
	   fprintf( pFile, "FNencode       = %d\n", ti->false_nencode);
	   fprintf( pFile, "NReduce        = %d\n", ti->nreduce);
	   fprintf( pFile, "FNReduce       = %d\n", ti->false_nreduce);
	   fprintf( pFile, "NCompare       = %d\n", ti->ncompare);
	   unsigned char timeBuf[TIME_BUF_SIZE];
     timeToString(&(ti->search_time), timeBuf);
	   fprintf( pFile, "Searching time = %s\n", timeBuf);
	   fclose(pFile);
}

int upgradeTable(TableInfo *ti)
{
    int lmp;
    if ((lmp = loadMainParameters(ti, 0) % 2) == 1){ 
       printf("Table not found!=\n");
       return 5; 
    }    
    if (checkParameters(ti)) return 4;
    if (ti->array_size < BOUND) return 1;
    unsigned char timeBuf[TIME_BUF_SIZE];
    
    randomize20(time(NULL));
  	initConsts(ti);  	
  	resetOutputValues(ti);
    if (lmp){
       lmp = 6; // since lmp will be returned
       printf("Description file not found.\n");
       printf("Creating description file. ");
       if (createDescriptionFile(ti)) // stores values into [table_name]2.SUF
          printf("FAILED\n");
       else
          printf("OK\n");    
    }
    int size1 = getTableSize(file_name, keyhash_sz);
    int size2 = ti->table_size;
  	void *pArray;
  	if ( (pArray = malloc(ti->array_size*keyhash_sz)) == NULL ){
           printf("Memory allocation error!");
           return 2;
  	}
  	
  	printf("Upgrading table %s\n", ti->table_name);
  	
    // Creates empty files file_name1... and ...1a.. 
  	FILE *pFile;
    if ((pFile = fopen(file_name1_tmp, "wb")) == NULL){  //resets file_name1
  	   printf("Error while upgrading table.\nFile cannot be opened %s", file_name1_tmp); 
  	   return 3;
    }
    fclose(pFile);
    if ((pFile = fopen(file_name1a_tmp, "wb")) == NULL){ // resets file_name1a
  	   printf("Error while upgrading table.\nFile cannot be opened %s", file_name1a_tmp); 
  	   return 3;
    }
    fclose(pFile);
    
    int counter = ti->limit1;
    if (counter <= 0) counter = 1;
    printf("Size of existing table: %d\n", size1);
    printf("Size to be added:       %d\n", size2);
    int ncol = 1; // any non zero value is ok; just to start the while cycle
    my_time timer_tmp;
    int ncol2 = 0;
    printf("Upgrading table...\n");
    now(&(ti->upgrade_time));
    while ((ncol)&&(counter--)){
          upgrade(ti, pArray, size1, size2, &ncol);
          size1 += size2 - ncol;
          size2 = ncol;
          ncol2 += global_ncollisions;
    }
    now(&timer_tmp);
    timeDifference(ti->upgrade_time, timer_tmp, &(ti->upgrade_time));
    printf("Table successfully upgraded!\n");
    timeToString(&(ti->upgrade_time), timeBuf);
    printf("Time of upgrading: %s seconds\n", timeBuf);
    
    ti->collisions = ncol2;
    ti->chains = ti->table_size - ncol;
    
    printf("%d chains were added to the table\n", ti->chains);
    printf("%d chains were dropped while upgrading.\n", ncol2);
    free(pArray);
  	return lmp;    
}

static void getItem(FILE *pFile, int i, void *outItem){
       fpos_t fpos = i*keyhash_sz;
       fsetpos(pFile, &fpos);
       fread(outItem, keyhash_sz, 1, pFile);
}

int searchTable(TableInfo *ti, const void *item)
{
    int lmp;
    my_time timer_tmp;
    if ((lmp = loadMainParameters(ti, 1) % 2) == 1){ 
       // lmp % 2 == 1 -> table does not exist
       printf("Error.\nFile cannot be opened %s", file_name); 
       return 2;
    }
    if (lmp) lmp = 1;
    resetOutputValues(ti);
    initConsts(ti);
  	FILE *pFile;
  	if ((pFile = fopen(file_name, "rb")) == NULL){
       printf("Error.\nFile cannot be opened %s", file_name); 
       return 2;
    }
    unsigned char keyhash[keyhash_sz]; // argument for lookup (inItem)
    unsigned char original_keyhash[keyhash_sz]; // to be compared with what is found in table
    unsigned char outItem[keyhash_sz]; // argument for lookup 
                                       // outItem[key] will be beginning of chain
    int i;
    // copy item to hash part of keyhash; initializing for lookup 
    // since for i=0 the cycle of j will not set the value of hash
    memcpy(keyhash + key_sz, (unsigned char *)item, hash_sz); 
    memcpy(original_keyhash+key_sz, (unsigned char *)item, hash_sz); 
    int j, k;
    int n = ti->chain_length;
    int pos;
    int items_found;
    now(&(ti->search_time));
    for (i=0; i<n; i++){
        // check (n-i)th column of chain (apply i * reduce(..n-i...), )
        for (j=0; j<i; j++){
            reduce(keyhash+key_sz, n-i+j, keyhash);
            encode(keyhash, keyhash+key_sz);
        }
        // lookup returns number of chains with matching endpoint found in table 
        if (items_found = lookup(pFile, ti->table_size, keyhash, outItem, &pos)){
           printf("items found %d\n", items_found);
           //check all chains with same endings
           for (k=0; k<items_found; k++){
               getItem(pFile, pos, outItem);
               for (j=1; j<n-i; j++){
                   encode(outItem, outItem+key_sz);
                   reduce(outItem+key_sz, j, outItem);     
               }
               // now have to check if hash(outItem[KEY])==item
               encode(outItem, outItem+key_sz);
/*               ti->nencode += (n-i);  
               printf("n %d, i %d", n, i);
               ti->nreduce += (n-i-1);*/
               if (compareHash(outItem, original_keyhash)){
                  // false alarm
                  ti->false_alarms++;
                  ti->false_nencode += (n-i);
                  ti->false_nreduce += (n-i-1);
               } else {
                  // key found
                  ti->nencode = ti->nreduce = (i*(i+1))/2;
                  ti->nencode += (n-i);
                  ti->nreduce += (n-i-1);
                  ti->ncompare += global_ncompare;
                  if ((ti->pKey = malloc(key_sz))==NULL){
                     now(&timer_tmp);
                     timeDifference(ti->search_time, timer_tmp, &(ti->search_time));
                     printf("Error. Key found, but memory allocation for key failed\n");
                     fclose(pFile);
                     return 1;
                  }              
                  now(&timer_tmp);
                  timeDifference(ti->search_time, timer_tmp, &(ti->search_time));
                  memcpy((unsigned char *)ti->pKey, (unsigned char *)outItem, key_sz); 
                  ti->row = pos;
                  ti->column = n-i-1;
                  fclose(pFile);
                  return 0;
               }
               pos++; // pos means position in table where chains were found
           } //for (k = ...))
        } //if (items_found = loookup(...))
        ti->ncompare += global_ncompare;
        memcpy((unsigned char*)(keyhash + key_sz), (unsigned char *)item, hash_sz); 
    }
    ti->nencode = ti->nreduce = ((n-1)*n)/2;
    now(&timer_tmp);
    timeDifference(ti->search_time, timer_tmp, &(ti->search_time));
    fclose(pFile);
    return 3 + lmp;
}

